"use strict";
/**
 * Borrow from https://github.com/excalidraw/excalidraw/blob/master/packages/excalidraw/change.ts#L399
 */
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.mutateElement = exports.ElementsChange = void 0;
const ecs_1 = require("@infinite-canvas-tutorial/ecs");
const util_1 = require("@antv/util");
const Delta_1 = require("./Delta");
const Snapshot_1 = require("./Snapshot");
const utils_1 = require("../utils");
class ElementsChange {
    static empty() {
        return ElementsChange.create(new Map(), new Map(), new Map());
    }
    static stripIrrelevantProps(partial) {
        const { id, version, versionNonce } = partial, strippedPartial = __rest(partial, ["id", "version", "versionNonce"]);
        return strippedPartial;
    }
    /**
     * Calculates the `Delta`s between the previous and next set of elements.
     *
     * @param prevElements - Map representing the previous state of elements.
     * @param nextElements - Map representing the next state of elements.
     *
     * @returns `ElementsChange` instance representing the `Delta` changes between the two sets of elements.
     */
    static calculate(prevElements, nextElements) {
        if (prevElements === nextElements) {
            return ElementsChange.empty();
        }
        const added = new Map();
        const removed = new Map();
        const updated = new Map();
        // this might be needed only in same edge cases, like during collab, when `isDeleted` elements get removed or when we (un)intentionally remove the elements
        for (const prevElement of prevElements.values()) {
            const nextElement = nextElements.get(prevElement.id);
            if (!nextElement) {
                const deleted = Object.assign(Object.assign({}, prevElement), { isDeleted: false });
                const inserted = { isDeleted: true };
                const delta = Delta_1.Delta.create(deleted, inserted, ElementsChange.stripIrrelevantProps);
                removed.set(prevElement.id, delta);
            }
        }
        for (const nextElement of nextElements.values()) {
            const prevElement = prevElements.get(nextElement.id);
            if (!prevElement) {
                const deleted = { isDeleted: true };
                const inserted = Object.assign(Object.assign({}, nextElement), { isDeleted: false });
                const delta = Delta_1.Delta.create(deleted, inserted, ElementsChange.stripIrrelevantProps);
                added.set(nextElement.id, delta);
                continue;
            }
            if (prevElement.versionNonce !== nextElement.versionNonce) {
                const delta = Delta_1.Delta.calculate(prevElement, nextElement, ElementsChange.stripIrrelevantProps);
                if (
                // making sure we don't get here some non-boolean values (i.e. undefined, null, etc.)
                typeof prevElement.isDeleted === 'boolean' &&
                    typeof nextElement.isDeleted === 'boolean' &&
                    prevElement.isDeleted !== nextElement.isDeleted) {
                    // notice that other props could have been updated as well
                    if (prevElement.isDeleted && !nextElement.isDeleted) {
                        added.set(nextElement.id, delta);
                    }
                    else {
                        removed.set(nextElement.id, delta);
                    }
                    continue;
                }
                // making sure there are at least some changes
                if (!Delta_1.Delta.isEmpty(delta)) {
                    updated.set(nextElement.id, delta);
                }
            }
        }
        return ElementsChange.create(added, removed, updated);
    }
    /**
     * Check for visible changes regardless of whether they were removed, added or updated.
     */
    static checkForVisibleDifference(element, partial) {
        if (element.isDeleted && partial.isDeleted !== false) {
            // when it's deleted and partial is not false, it cannot end up with a visible change
            return false;
        }
        if (element.isDeleted && partial.isDeleted === false) {
            // when we add an element, it results in a visible change
            return true;
        }
        if (element.isDeleted === false && partial.isDeleted) {
            // when we remove an element, it results in a visible change
            return true;
        }
        // check for any difference on a visible element
        return Delta_1.Delta.isRightDifferent(element, partial);
    }
    static applyDelta(element, delta, flags = {
        // by default we don't care about about the flags
        containsVisibleDifference: true,
        containsZindexDifference: true,
    }) {
        const directlyApplicablePartial = __rest(delta.inserted, []);
        // if (
        //   delta.deleted.boundElements?.length ||
        //   delta.inserted.boundElements?.length
        // ) {
        //   const mergedBoundElements = Delta.mergeArrays(
        //     element.boundElements,
        //     delta.inserted.boundElements,
        //     delta.deleted.boundElements,
        //     (x) => x.id,
        //   );
        //   Object.assign(directlyApplicablePartial, {
        //     boundElements: mergedBoundElements,
        //   });
        // }
        // if (isImageElement(element)) {
        //   const _delta = delta as Delta<ElementPartial<ExcalidrawImageElement>>;
        //   // we want to override `crop` only if modified so that we don't reset
        //   // when undoing/redoing unrelated change
        //   if (_delta.deleted.crop || _delta.inserted.crop) {
        //     Object.assign(directlyApplicablePartial, {
        //       // apply change verbatim
        //       crop: _delta.inserted.crop ?? null,
        //     });
        //   }
        // }
        if (!flags.containsVisibleDifference) {
            // strip away fractional as even if it would be different, it doesn't have to result in visible change
            const { index } = directlyApplicablePartial, rest = __rest(directlyApplicablePartial, ["index"]);
            const containsVisibleDifference = ElementsChange.checkForVisibleDifference(element, rest);
            flags.containsVisibleDifference = containsVisibleDifference;
        }
        if (!flags.containsZindexDifference) {
            flags.containsZindexDifference =
                delta.deleted.index !== delta.inserted.index;
        }
        return (0, Snapshot_1.newElementWith)(element, directlyApplicablePartial);
    }
    static create(added, removed, updated, options = { shouldRedistribute: false }) {
        let change;
        if (options.shouldRedistribute) {
            const nextAdded = new Map();
            const nextRemoved = new Map();
            const nextUpdated = new Map();
            const deltas = [...added, ...removed, ...updated];
            for (const [id, delta] of deltas) {
                if (this.satisfiesAddition(delta)) {
                    nextAdded.set(id, delta);
                }
                else if (this.satisfiesRemoval(delta)) {
                    nextRemoved.set(id, delta);
                }
                else {
                    nextUpdated.set(id, delta);
                }
            }
            change = new ElementsChange(nextAdded, nextRemoved, nextUpdated);
        }
        else {
            change = new ElementsChange(added, removed, updated);
        }
        return change;
    }
    constructor(added, removed, updated) {
        this.added = added;
        this.removed = removed;
        this.updated = updated;
    }
    inverse() {
        const inverseInternal = (deltas) => {
            const inversedDeltas = new Map();
            for (const [id, delta] of deltas.entries()) {
                inversedDeltas.set(id, Delta_1.Delta.create(delta.inserted, delta.deleted));
            }
            return inversedDeltas;
        };
        const added = inverseInternal(this.added);
        const removed = inverseInternal(this.removed);
        const updated = inverseInternal(this.updated);
        // notice we inverse removed with added not to break the invariants
        return ElementsChange.create(removed, added, updated);
    }
    /**
     * Update delta/s based on the existing elements.
     *
     * @param elements current elements
     * @param modifierOptions defines which of the delta (`deleted` or `inserted`) will be updated
     * @returns new instance with modified delta/s
     */
    applyLatestChanges(elements) {
        const modifier = (element) => (partial) => {
            // (element: OrderedSerializedNode) => (partial: ElementPartial) => {
            const latestPartial = {};
            for (const key of Object.keys(partial)) {
                // do not update following props:
                // - `boundElements`, as it is a reference value which is postprocessed to contain only deleted/inserted keys
                switch (key) {
                    // case 'boundElements':
                    //   latestPartial[key] = partial[key];
                    //   break;
                    default:
                        latestPartial[key] = element[key];
                }
            }
            return latestPartial;
        };
        const applyLatestChangesInternal = (deltas) => {
            const modifiedDeltas = new Map();
            for (const [id, delta] of deltas.entries()) {
                const existingElement = elements.get(id);
                if (existingElement) {
                    const modifiedDelta = Delta_1.Delta.create(delta.deleted, delta.inserted, modifier(existingElement), 'inserted');
                    modifiedDeltas.set(id, modifiedDelta);
                }
                else {
                    modifiedDeltas.set(id, delta);
                }
            }
            return modifiedDeltas;
        };
        const added = applyLatestChangesInternal(this.added);
        const removed = applyLatestChangesInternal(this.removed);
        const updated = applyLatestChangesInternal(this.updated);
        return ElementsChange.create(added, removed, updated, {
            shouldRedistribute: true, // redistribute the deltas as `isDeleted` could have been updated
        });
    }
    applyTo(elements, snapshot) {
        let nextElements = new Map(elements);
        let changedElements;
        const flags = {
            containsVisibleDifference: false,
            containsZindexDifference: false,
        };
        const applyDeltas = ElementsChange.createApplier(nextElements, snapshot, flags);
        const addedElements = applyDeltas(this.added);
        const removedElements = applyDeltas(this.removed);
        const updatedElements = applyDeltas(this.updated);
        console.log('addedElements', addedElements);
        console.log('removedElements', removedElements);
        console.log('updatedElements', updatedElements);
        const affectedElements = this.resolveConflicts(elements, nextElements);
        // TODO: #7348 validate elements semantically and syntactically the changed elements, in case they would result data integrity issues
        changedElements = new Map([
            ...addedElements,
            ...removedElements,
            ...updatedElements,
            ...affectedElements,
        ]);
        // the following reorder performs also mutations, but only on new instances of changed elements
        // (unless something goes really bad and it fallbacks to fixing all invalid indices)
        // nextElements = ElementsChange.reorderElements(
        //   nextElements,
        //   changedElements,
        //   flags,
        // );
        return [nextElements, flags.containsVisibleDifference];
    }
    isEmpty() {
        return (this.added.size === 0 &&
            this.removed.size === 0 &&
            this.updated.size === 0);
    }
    /**
     * Resolves conflicts for all previously added, removed and updated elements.
     * Updates the previous deltas with all the changes after conflict resolution.
     *
     * @returns all elements affected by the conflict resolution
     */
    resolveConflicts(prevElements, nextElements) {
        const nextAffectedElements = new Map();
        const updater = (element, updates) => {
            const nextElement = nextElements.get(element.id); // only ever modify next element!
            if (!nextElement) {
                return;
            }
            let affectedElement;
            if (prevElements.get(element.id) === nextElement) {
                // create the new element instance in case we didn't modify the element yet
                // so that we won't end up in an incosistent state in case we would fail in the middle of mutations
                affectedElement = (0, Snapshot_1.newElementWith)(nextElement, updates);
            }
            else {
                console.log('mutateElement', nextElement, updates);
                // affectedElement = mutateElement(
                //   nextElement,
                //   updates as ElementUpdate<OrderedSerializedNode>,
                // ) as OrderedSerializedNode;
            }
            nextAffectedElements.set(affectedElement.id, affectedElement);
            nextElements.set(affectedElement.id, affectedElement);
        };
        // // removed delta is affecting the bindings always, as all the affected elements of the removed elements need to be unbound
        // for (const [id] of this.removed) {
        //   ElementsChange.unbindAffected(prevElements, nextElements, id, updater);
        // }
        // // added delta is affecting the bindings always, all the affected elements of the added elements need to be rebound
        // for (const [id] of this.added) {
        //   ElementsChange.rebindAffected(prevElements, nextElements, id, updater);
        // }
        // updated delta is affecting the binding only in case it contains changed binding or bindable property
        // .filter(([_, delta]) =>
        //   Object.keys({ ...delta.deleted, ...delta.inserted }).find((prop) =>
        //     bindingProperties.has(prop as BindingProp | BindableProp),
        //   ),
        // )
        for (const [id] of Array.from(this.updated)) {
            const updatedElement = nextElements.get(id);
            if (!updatedElement || updatedElement.isDeleted) {
                // skip fixing bindings for updates on deleted elements
                continue;
            }
            updater(updatedElement, updatedElement);
            // ElementsChange.rebindAffected(prevElements, nextElements, id, updater);
        }
        // filter only previous elements, which were now affected
        const prevAffectedElements = new Map(Array.from(prevElements).filter(([id]) => nextAffectedElements.has(id)));
        // calculate complete deltas for affected elements, and assign them back to all the deltas
        // technically we could do better here if perf. would become an issue
        const { added, removed, updated } = ElementsChange.calculate(prevAffectedElements, nextAffectedElements);
        for (const [id, delta] of added) {
            this.added.set(id, delta);
        }
        for (const [id, delta] of removed) {
            this.removed.set(id, delta);
        }
        for (const [id, delta] of updated) {
            this.updated.set(id, delta);
        }
        return nextAffectedElements;
    }
}
exports.ElementsChange = ElementsChange;
ElementsChange.satisfiesAddition = ({ deleted, inserted, }) => 
// dissallowing added as "deleted", which could cause issues when resolving conflicts
deleted.isDeleted === true && !inserted.isDeleted;
ElementsChange.satisfiesRemoval = ({ deleted, inserted, }) => !deleted.isDeleted && inserted.isDeleted === true;
ElementsChange.satisfiesUpdate = ({ deleted, inserted, }) => !!deleted.isDeleted === !!inserted.isDeleted;
ElementsChange.createGetter = (elements, snapshot, flags) => (id, partial) => {
    let element = elements.get(id);
    if (!element) {
        // always fallback to the local snapshot, in cases when we cannot find the element in the elements array
        element = snapshot.get(id);
        if (element) {
            // as the element was brought from the snapshot, it automatically results in a possible zindex difference
            flags.containsZindexDifference = true;
            // as the element was force deleted, we need to check if adding it back results in a visible change
            if (partial.isDeleted === false ||
                (partial.isDeleted !== true && element.isDeleted === false)) {
                flags.containsVisibleDifference = true;
            }
        }
    }
    return element;
};
ElementsChange.createApplier = (nextElements, snapshot, flags) => {
    const getElement = ElementsChange.createGetter(nextElements, snapshot, flags);
    return (deltas) => Array.from(deltas.entries()).reduce((acc, [id, delta]) => {
        const element = getElement(id, delta.inserted);
        if (element) {
            const newElement = ElementsChange.applyDelta(element, delta, flags);
            nextElements.set(newElement.id, newElement);
            acc.set(newElement.id, newElement);
        }
        return acc;
    }, new Map());
};
// This function tracks updates of text elements for the purposes for collaboration.
// The version is used to compare updates when more than one user is working in
// the same drawing. Note: this will trigger the component to update. Make sure you
// are calling it either from a React event handler or within unstable_batchedUpdates().
const mutateElement = (entity, element, updates) => {
    let didChange = false;
    for (const key in updates) {
        const value = updates[key];
        if (typeof value !== 'undefined') {
            element[key] = value;
            didChange = true;
        }
    }
    if (!didChange) {
        return element;
    }
    const { name, visibility } = updates;
    const { fill, stroke, strokeWidth, strokeCap, strokeJoin, strokeAlignment, fontSize, width, height, x, y, } = updates;
    if (!(0, util_1.isNil)(name)) {
        entity.write(ecs_1.Name).value = name;
    }
    if (!(0, util_1.isNil)(visibility)) {
        entity.write(ecs_1.Visibility).value = visibility;
    }
    if (!(0, util_1.isNil)(fill)) {
        entity.write(ecs_1.FillSolid).value = fill;
    }
    if (!(0, util_1.isNil)(stroke)) {
        (0, utils_1.createOrSetComponent)(entity, ecs_1.Stroke, { color: stroke });
    }
    if (!(0, util_1.isNil)(strokeWidth)) {
        (0, utils_1.createOrSetComponent)(entity, ecs_1.Stroke, { width: strokeWidth });
    }
    if (!(0, util_1.isNil)(strokeCap)) {
        (0, utils_1.createOrSetComponent)(entity, ecs_1.Stroke, { linecap: strokeCap });
    }
    if (!(0, util_1.isNil)(strokeJoin)) {
        (0, utils_1.createOrSetComponent)(entity, ecs_1.Stroke, { linejoin: strokeJoin });
    }
    if (!(0, util_1.isNil)(strokeAlignment)) {
        (0, utils_1.createOrSetComponent)(entity, ecs_1.Stroke, { alignment: strokeAlignment });
    }
    if (!(0, util_1.isNil)(fontSize)) {
        entity.write(ecs_1.Text).fontSize = fontSize;
    }
    if (!(0, util_1.isNil)(width)) {
        entity.write(ecs_1.Rect).width = width;
    }
    if (!(0, util_1.isNil)(height)) {
        entity.write(ecs_1.Rect).height = height;
    }
    if (!(0, util_1.isNil)(x)) {
        entity.write(ecs_1.Rect).x = x;
    }
    if (!(0, util_1.isNil)(y)) {
        entity.write(ecs_1.Rect).y = y;
    }
    if ((0, util_1.isNil)(element.version)) {
        element.version = 0;
    }
    element.version++;
    element.versionNonce = (0, ecs_1.randomInteger)();
    element.updated = (0, utils_1.getUpdatedTimestamp)();
    return element;
};
exports.mutateElement = mutateElement;
//# sourceMappingURL=ElementsChange.js.map